module Bag(Bag, assocs, insert, empty, pop, count, fromList, toSet, size, intersection, toAssoc) where

import Data.Maybe
import Data.List hiding(insert)
import qualified Data.Map as M
import qualified Data.Set as S

newtype Bag a = Bag (M.Map a Int)
  deriving (Eq, Ord)

assocs :: (Ord a) => Bag a -> [(a, Int)]
assocs (Bag m) = M.assocs m

insert :: (Ord a) => a -> Bag a -> Bag a
insert x (Bag m) = Bag $ M.insertWith (+) x 1 m

empty :: Bag a
empty = Bag M.empty

pop :: (Ord a) => Bag a -> Maybe (a, Bag a)
pop (Bag m) = do
  ((x, c), r) <- M.minViewWithKey m 
  if c == 1
    then return (x, Bag r)
    else return (x, Bag $ M.insert x (c-1) r)

count :: (Ord a) => a -> Bag a -> Int
count x (Bag m) = fromMaybe 0 $ M.lookup x m

fromList :: (Ord a) => [a] -> Bag a
fromList xs = foldl' (flip insert) empty xs

toSet :: (Ord a) => Bag a -> S.Set a
toSet (Bag m) = M.keysSet m

size :: (Ord a) => Bag a -> Int
size (Bag m) = M.fold (+) 0 m

intersection :: (Ord a) => Bag a -> Bag a -> Bag a
intersection (Bag x) (Bag y) = Bag $ M.intersectionWith min x y

toAssoc :: (Ord a) => Bag a -> [(a, Int)]
toAssoc (Bag m) = M.toList m
