{-# LANGUAGE ForeignFunctionInterface #-}
module CXXMatch(match, quietMatch) where

import Char
import Swords
import qualified Match

import System.IO.SSC.Haskell98 as SSC
import System.IO.SSC.Haskell98.Base(withHandle, haFd) -- hack: violating the abstraction barrier

import Control.Monad
import qualified Data.Map as M
import Foreign.Ptr
import Foreign.C.Types
import Foreign.Marshal.Array
import Foreign.C.String
import Foreign.Storable
import System.Posix.Types(Fd(..))
import System.IO.Unsafe

type SRef = Ptr ()

match :: SSC.Handle -> Character -> Character -> IO GameResult
match out x y = do
  hFlush out
  Just (Fd fd) <- withHandle out $ return . haFd
  mr <- matchGeneral fd x y
  case mr of
    Just r -> return r
    Nothing -> do
      hPutStrLn out "\n*** シミュレーション失敗。旧エンジンで再試行 ***\n"
      let (gr, msgs, _) = Match.match x y
      hPutStr out $ unlines msgs
      grWinner gr `seq` return gr

quietMatch :: Character -> Character -> GameResult
quietMatch x y = case unsafePerformIO $ matchGeneral (-1) x y of
  Just r -> r
  Nothing -> case Match.match x y of
    (r, _, _) -> r

matchGeneral :: CInt -> Character -> Character -> IO (Maybe GameResult)
matchGeneral fd x y =
  withSwords (chSwords x) $ \sx lenx ->
  withSwords (chSwords y) $ \sy leny ->
  withCWString (chName x) $ \nx ->
  withCWString (chName y) $ \ny -> do
    r <- c_match_for_haskell fd
            nx (fi $ chHp x) (fi $ chAtk x) (fi $ chDef x) (fi $ chSpd x) sx lenx
            ny (fi $ chHp y) (fi $ chAtk y) (fi $ chDef y) (fi $ chSpd y) sy leny
    if r == 3
      then return Nothing
      else return $ Just GR{ grWinner = [Just Team0, Just Team1, Nothing] !! (fi r) }

fi :: (Integral a, Num b) => a -> b
fi = fromIntegral

withSwords :: [Sword] -> (Ptr SRef -> CInt -> IO a) -> IO a
withSwords ss action = do
  srs <- mapM convertSword ss
  withArrayLen srs $ \len ptr -> action ptr (fi len)

convertSword :: Sword -> IO SRef
convertSword sw = case M.lookup sw swordMap of
  Just sr -> sr
  Nothing -> error $ "convertSword: not found: " ++ show sw

swordMap :: M.Map Sword (IO SRef)
swordMap = M.fromList tbl
  where
    tbl = map csword tbl_constant ++ concatMap psword tbl_parametric
    csword (sw, ptr) = (sw, peek ptr)
    psword (gs, gfun) = map (\n -> (gs n, gfun $ fi n)) [0..20]

tbl_constant :: [(Sword, Ptr SRef)]
tbl_constant = 
  [ (slash, c_s_slash)
  , (rapid, c_s_rapid)
  , (heavy, c_s_heavy)
  , (shock, c_s_shock)
  , (poison, c_s_poison)
  , (absolute, c_s_absolute)
  , (magical, c_s_magical)
  , (death, c_s_death)
  , (heat, c_s_heat)
  , (protect, c_s_protect)
  , (accel, c_s_accel)
  , (shield, c_s_shield)
  , (mirror, c_s_mirror)
  , (life, c_s_life)

  , (nuru, c_s_null)
  , (spark, c_s_spark)
  , (firecracker, c_s_firecracker)
  , (eruption, c_s_eruption)
  , (devour, c_s_devour)
  , (disaster, c_s_disaster)
  , (perish, c_s_perish)
  , (finale, c_s_finale)
  , (combo, c_s_combo)
  , (wave, c_s_wave)
  , (dance, c_s_dance)
  , (scatter, c_s_scatter)
  , (toxic, c_s_toxic)
  , (abortion, c_s_abortion)
  , (destroy, c_s_destroy)
  , (battle, c_s_battle)
  , (confusion, c_s_confusion)
  , (bloodsuck, c_s_bloodsuck)
  , (robbery, c_s_robbery)
  , (dark, c_s_dark)
  , (antimagic, c_s_antimagic)
  , (overweight, c_s_overweight)
  , (solid, c_s_solid)
  , (sonic, c_s_sonic)
  , (light, c_s_light)
  , (cool, c_s_cool)
  , (decay, c_s_decay)
  , (decel, c_s_decel)
  , (gentle, c_s_gentle)
  , (flashfire, c_s_flashfire)
  , (alert, c_s_alert)
  , (evasion, c_s_evasion)
  , (escutcheon, c_s_escutcheon)
  , (amnion, c_s_amnion)
  , (barrier, c_s_barrier)
  , (soul, c_s_soul)
  , (push, c_s_push)
  , (drive, c_s_drive)
  , (wriggle, c_s_wriggle)
  , (attraction, c_s_attraction)
  , (first, c_s_first)
  , (second, c_s_second)
  , (communication, c_s_communication)
  , (chant, c_s_chant)
  , (prayer, c_s_prayer)
  , (prospect, c_s_prospect)
  , (dragonize, c_s_dragonize)
  , (limit, c_s_limit)
  , (suppression, c_s_suppression)
  , (vibration, c_s_vibration)
  , (toxicant, c_s_toxicant)
  , (diffuse, c_s_diffuse)
  , (train, c_s_train)
  , (fictional, c_s_fictional)
  , (multi, c_s_multi)
  , (replica, c_s_replica)
  , (reverb, c_s_reverb)
  , (echo, c_s_echo)
  , (bubble, c_s_bubble)
  , (spirit, c_s_spirit)
  , (fetal, c_s_fetal)
  , (sawhair, c_s_sawhair)
  , (sawa, c_s_sawa)
  , (purple, c_s_purple)
  , (holand, c_s_holand)
  , (servant, c_s_servant)
  , (leg, c_s_leg)
  , (dog, c_s_dog)
  , (ko, c_s_ko)
  , (yoshi, c_s_yoshi)
  , (holy, c_s_holy)
  , (geki, c_s_geki)
  , (emperor, c_s_emperor)
  , (bone, c_s_bone)
  , (princess, c_s_princess)
  , (summon, c_s_summon)
  , (age, c_s_age)

  , (recoveryA, c_s_recoveryA)
  , (perishA, c_s_perishA)
  , (rageA, c_s_rageA)
  , (fang, c_s_fang)
  , (two, c_s_two)
  , (three, c_s_three)
  , (four, c_s_four)
  , (five, c_s_five)
  , (six, c_s_six)
  , (seven, c_s_seven)
  , (abortionA, c_s_abortionA)
  , (coolA, c_s_coolA)
  , (decayA, c_s_decayA)
  , (decelA, c_s_decelA)
  , (darkA, c_s_darkA)
  , (antimagicA, c_s_antimagicA)
  , (pushA, c_s_pushA)
  , (attractionA, c_s_attractionA)
  , (healA, c_s_healA)
  , (chantA, c_s_chantA)
  , (rogation, c_s_rogation)
  , (ritualA, c_s_ritualA)
  , (blaze, c_s_blaze)
  , (deriveA, c_s_deriveA)
  , (slow, c_s_slow)
  , (danger, c_s_danger)
  , (recurrenceA, c_s_recurrenceA)
  , (resonanceA, c_s_resonanceA)
  , (ageA, c_s_ageA)
  , (mkbubble, c_s_mkbubble)
  , (mkfence, c_s_mkfence)
  , (mkgate, c_s_mkgate)
  , (mkwall, c_s_mkwall)
  , (mkice, c_s_mkice)
  , (mkmist, c_s_mkmist)
  , (mktear, c_s_mktear)
  ]

tbl_parametric :: [(Int -> Sword, CInt -> IO SRef)]
tbl_parametric =
  [ (recovery, c_s_recovery)
  , (rage, c_s_rage)
  , (heal, c_s_heal)
  , (ritual, c_s_ritual)
  , (delay, c_s_delay)
  , (sheathe, c_s_sheathe)
  , (derive, c_s_derive)
  , (recurrence, c_s_recurrence)
  , (resonance, c_s_resonance)

  , (victory, c_s_victory)
  ]

foreign import ccall "haskell_interface.h m100sim_match_for_haskell"
  c_match_for_haskell 
    :: CInt
    -> CWString -> CInt -> CInt -> CInt -> CInt -> Ptr SRef -> CInt
    -> CWString -> CInt -> CInt -> CInt -> CInt -> Ptr SRef -> CInt
    -> IO CInt
foreign import ccall "match.h &m100sim_s_slash" c_s_slash :: Ptr SRef
foreign import ccall "match.h &m100sim_s_rapid" c_s_rapid :: Ptr SRef
foreign import ccall "match.h &m100sim_s_heavy" c_s_heavy :: Ptr SRef
foreign import ccall "match.h &m100sim_s_shock" c_s_shock :: Ptr SRef
foreign import ccall "match.h &m100sim_s_poison" c_s_poison :: Ptr SRef
foreign import ccall "match.h &m100sim_s_absolute" c_s_absolute :: Ptr SRef
foreign import ccall "match.h &m100sim_s_magical" c_s_magical :: Ptr SRef
foreign import ccall "match.h &m100sim_s_death" c_s_death :: Ptr SRef
foreign import ccall "match.h &m100sim_s_heat" c_s_heat :: Ptr SRef
foreign import ccall "match.h &m100sim_s_protect" c_s_protect :: Ptr SRef
foreign import ccall "match.h &m100sim_s_accel" c_s_accel :: Ptr SRef
foreign import ccall "match.h &m100sim_s_shield" c_s_shield :: Ptr SRef
foreign import ccall "match.h &m100sim_s_mirror" c_s_mirror :: Ptr SRef
foreign import ccall "match.h &m100sim_s_life" c_s_life :: Ptr SRef
foreign import ccall "match.h m100sim_s_recovery" c_s_recovery :: CInt -> IO SRef

foreign import ccall "match.h &m100sim_s_null" c_s_null :: Ptr SRef
foreign import ccall "match.h &m100sim_s_spark" c_s_spark :: Ptr SRef
foreign import ccall "match.h &m100sim_s_firecracker" c_s_firecracker :: Ptr SRef
foreign import ccall "match.h &m100sim_s_eruption" c_s_eruption :: Ptr SRef
foreign import ccall "match.h &m100sim_s_devour" c_s_devour :: Ptr SRef
foreign import ccall "match.h &m100sim_s_disaster" c_s_disaster :: Ptr SRef
foreign import ccall "match.h &m100sim_s_perish" c_s_perish :: Ptr SRef
foreign import ccall "match.h &m100sim_s_finale" c_s_finale :: Ptr SRef
foreign import ccall "match.h &m100sim_s_combo" c_s_combo :: Ptr SRef
foreign import ccall "match.h &m100sim_s_wave" c_s_wave :: Ptr SRef
foreign import ccall "match.h &m100sim_s_dance" c_s_dance :: Ptr SRef
foreign import ccall "match.h &m100sim_s_scatter" c_s_scatter :: Ptr SRef
foreign import ccall "match.h &m100sim_s_toxic" c_s_toxic :: Ptr SRef
foreign import ccall "match.h &m100sim_s_abortion" c_s_abortion :: Ptr SRef
foreign import ccall "match.h &m100sim_s_destroy" c_s_destroy :: Ptr SRef
foreign import ccall "match.h &m100sim_s_battle" c_s_battle :: Ptr SRef
foreign import ccall "match.h &m100sim_s_confusion" c_s_confusion :: Ptr SRef
foreign import ccall "match.h &m100sim_s_bloodsuck" c_s_bloodsuck :: Ptr SRef
foreign import ccall "match.h &m100sim_s_robbery" c_s_robbery :: Ptr SRef
foreign import ccall "match.h &m100sim_s_dark" c_s_dark :: Ptr SRef
foreign import ccall "match.h &m100sim_s_antimagic" c_s_antimagic :: Ptr SRef
foreign import ccall "match.h &m100sim_s_overweight" c_s_overweight :: Ptr SRef
foreign import ccall "match.h &m100sim_s_solid" c_s_solid :: Ptr SRef
foreign import ccall "match.h &m100sim_s_sonic" c_s_sonic :: Ptr SRef
foreign import ccall "match.h &m100sim_s_light" c_s_light :: Ptr SRef
foreign import ccall "match.h m100sim_s_rage" c_s_rage :: CInt -> IO SRef
foreign import ccall "match.h &m100sim_s_cool" c_s_cool :: Ptr SRef
foreign import ccall "match.h &m100sim_s_decay" c_s_decay :: Ptr SRef
foreign import ccall "match.h &m100sim_s_decel" c_s_decel :: Ptr SRef
foreign import ccall "match.h &m100sim_s_gentle" c_s_gentle :: Ptr SRef
foreign import ccall "match.h &m100sim_s_flashfire" c_s_flashfire :: Ptr SRef
foreign import ccall "match.h &m100sim_s_alert" c_s_alert :: Ptr SRef
foreign import ccall "match.h &m100sim_s_evasion" c_s_evasion :: Ptr SRef
foreign import ccall "match.h &m100sim_s_escutcheon" c_s_escutcheon :: Ptr SRef
foreign import ccall "match.h &m100sim_s_amnion" c_s_amnion :: Ptr SRef
foreign import ccall "match.h &m100sim_s_barrier" c_s_barrier :: Ptr SRef
foreign import ccall "match.h &m100sim_s_soul" c_s_soul :: Ptr SRef
foreign import ccall "match.h &m100sim_s_push" c_s_push :: Ptr SRef
foreign import ccall "match.h &m100sim_s_drive" c_s_drive :: Ptr SRef
foreign import ccall "match.h &m100sim_s_wriggle" c_s_wriggle :: Ptr SRef
foreign import ccall "match.h &m100sim_s_attraction" c_s_attraction :: Ptr SRef
foreign import ccall "match.h &m100sim_s_first" c_s_first :: Ptr SRef
foreign import ccall "match.h &m100sim_s_second" c_s_second :: Ptr SRef
foreign import ccall "match.h &m100sim_s_communication" c_s_communication :: Ptr SRef
foreign import ccall "match.h m100sim_s_heal" c_s_heal :: CInt -> IO SRef
foreign import ccall "match.h &m100sim_s_chant" c_s_chant :: Ptr SRef
foreign import ccall "match.h &m100sim_s_prayer" c_s_prayer :: Ptr SRef
foreign import ccall "match.h m100sim_s_ritual" c_s_ritual :: CInt -> IO SRef
foreign import ccall "match.h &m100sim_s_prospect" c_s_prospect :: Ptr SRef
foreign import ccall "match.h &m100sim_s_dragonize" c_s_dragonize :: Ptr SRef
foreign import ccall "match.h &m100sim_s_limit" c_s_limit :: Ptr SRef
foreign import ccall "match.h &m100sim_s_suppression" c_s_suppression :: Ptr SRef
foreign import ccall "match.h m100sim_s_delay" c_s_delay :: CInt -> IO SRef
foreign import ccall "match.h m100sim_s_sheathe" c_s_sheathe :: CInt -> IO SRef
foreign import ccall "match.h m100sim_s_derive" c_s_derive :: CInt -> IO SRef
foreign import ccall "match.h &m100sim_s_vibration" c_s_vibration :: Ptr SRef
foreign import ccall "match.h &m100sim_s_toxicant" c_s_toxicant :: Ptr SRef
foreign import ccall "match.h &m100sim_s_diffuse" c_s_diffuse :: Ptr SRef
foreign import ccall "match.h &m100sim_s_train" c_s_train :: Ptr SRef
foreign import ccall "match.h m100sim_s_recurrence" c_s_recurrence :: CInt -> IO SRef
foreign import ccall "match.h m100sim_s_resonance" c_s_resonance :: CInt -> IO SRef
foreign import ccall "match.h &m100sim_s_fictional" c_s_fictional :: Ptr SRef
foreign import ccall "match.h &m100sim_s_multi" c_s_multi :: Ptr SRef
foreign import ccall "match.h &m100sim_s_replica" c_s_replica :: Ptr SRef
foreign import ccall "match.h &m100sim_s_reverb" c_s_reverb :: Ptr SRef
foreign import ccall "match.h &m100sim_s_echo" c_s_echo :: Ptr SRef
foreign import ccall "match.h &m100sim_s_bubble" c_s_bubble :: Ptr SRef
foreign import ccall "match.h &m100sim_s_spirit" c_s_spirit :: Ptr SRef
foreign import ccall "match.h &m100sim_s_fetal" c_s_fetal :: Ptr SRef
foreign import ccall "match.h &m100sim_s_sawhair" c_s_sawhair :: Ptr SRef
foreign import ccall "match.h &m100sim_s_sawa" c_s_sawa :: Ptr SRef
foreign import ccall "match.h &m100sim_s_purple" c_s_purple :: Ptr SRef
foreign import ccall "match.h &m100sim_s_holand" c_s_holand :: Ptr SRef
foreign import ccall "match.h &m100sim_s_servant" c_s_servant :: Ptr SRef
foreign import ccall "match.h &m100sim_s_leg" c_s_leg :: Ptr SRef
foreign import ccall "match.h &m100sim_s_dog" c_s_dog :: Ptr SRef
foreign import ccall "match.h &m100sim_s_ko" c_s_ko :: Ptr SRef
foreign import ccall "match.h &m100sim_s_yoshi" c_s_yoshi :: Ptr SRef
foreign import ccall "match.h &m100sim_s_holy" c_s_holy :: Ptr SRef
foreign import ccall "match.h &m100sim_s_geki" c_s_geki :: Ptr SRef
foreign import ccall "match.h &m100sim_s_emperor" c_s_emperor :: Ptr SRef
foreign import ccall "match.h &m100sim_s_bone" c_s_bone :: Ptr SRef
foreign import ccall "match.h &m100sim_s_princess" c_s_princess :: Ptr SRef
foreign import ccall "match.h &m100sim_s_summon" c_s_summon :: Ptr SRef
foreign import ccall "match.h &m100sim_s_age" c_s_age :: Ptr SRef

foreign import ccall "match.h &m100sim_s_recoveryA" c_s_recoveryA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_perishA" c_s_perishA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_rageA" c_s_rageA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_fang" c_s_fang :: Ptr SRef
foreign import ccall "match.h &m100sim_s_two" c_s_two :: Ptr SRef
foreign import ccall "match.h &m100sim_s_three" c_s_three :: Ptr SRef
foreign import ccall "match.h &m100sim_s_four" c_s_four :: Ptr SRef
foreign import ccall "match.h &m100sim_s_five" c_s_five :: Ptr SRef
foreign import ccall "match.h &m100sim_s_six" c_s_six :: Ptr SRef
foreign import ccall "match.h &m100sim_s_seven" c_s_seven :: Ptr SRef
foreign import ccall "match.h &m100sim_s_abortionA" c_s_abortionA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_coolA" c_s_coolA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_decayA" c_s_decayA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_decelA" c_s_decelA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_darkA" c_s_darkA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_antimagicA" c_s_antimagicA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_pushA" c_s_pushA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_attractionA" c_s_attractionA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_healA" c_s_healA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_chantA" c_s_chantA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_rogation" c_s_rogation :: Ptr SRef
foreign import ccall "match.h &m100sim_s_ritualA" c_s_ritualA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_blaze" c_s_blaze :: Ptr SRef
foreign import ccall "match.h &m100sim_s_deriveA" c_s_deriveA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_slow" c_s_slow :: Ptr SRef
foreign import ccall "match.h &m100sim_s_danger" c_s_danger :: Ptr SRef
foreign import ccall "match.h &m100sim_s_recurrenceA" c_s_recurrenceA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_resonanceA" c_s_resonanceA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_ageA" c_s_ageA :: Ptr SRef
foreign import ccall "match.h &m100sim_s_mkbubble" c_s_mkbubble :: Ptr SRef
foreign import ccall "match.h &m100sim_s_mkfence" c_s_mkfence :: Ptr SRef
foreign import ccall "match.h &m100sim_s_mkgate" c_s_mkgate :: Ptr SRef
foreign import ccall "match.h &m100sim_s_mkwall" c_s_mkwall :: Ptr SRef
foreign import ccall "match.h &m100sim_s_mkice" c_s_mkice :: Ptr SRef
foreign import ccall "match.h &m100sim_s_mkmist" c_s_mkmist :: Ptr SRef
foreign import ccall "match.h &m100sim_s_mktear" c_s_mktear :: Ptr SRef

foreign import ccall "match.h m100sim_s_victory" c_s_victory :: CInt -> IO SRef
