-- #hide
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental
   
   rawGetBufB\/rawPutBufB interface for fds.
-}
module System.IO.SSC.Internal.Fd
  ( module System.IO.Fd
  , fdGetBufB
  , fdPutBufB
  , fdRecvB
  , fdSendB
  ) where

import Foreign.Ptr
import System.IO.Fd
import System.IO.SSC.Classes

fdGetBufB, fdPutBufB, fdRecvB, fdSendB :: Fd -> BlockingMode -> Ptr a -> Int -> IO (Int, Bool)
fdGetBufB = toB fdGetBuf fdGetBufNonBlocking
fdPutBufB = toB fdPutBuf fdPutBufNonBlocking
fdRecvB = toB fdRecv fdRecvNonBlocking
fdSendB = toB fdSend fdSendNonBlocking

toB
  :: (Fd -> Ptr a -> Int -> IO Int)
  -> (Fd -> Ptr a -> Int -> IO (Maybe Int))
  -> Fd
  -> BlockingMode
  -> Ptr a
  -> Int
  -> IO (Int, Bool)
toB plain nb fd bmode buf 0 = return (0, False)
toB plain nb fd bmode buf size = case bmode of
  NonBlockingPreferred -> do
    r <- plain fd buf size
    return (r, r == 0)
  NonBlocking -> do
    r <- nb fd buf size
    case r of
      Nothing -> return (0, False)
      Just 0 -> return (0, True)
      Just n -> return (n, False)
  Blocking -> loop 0
    where
      loop n
        | n == size = return (n, False)
        | otherwise = do
          r <- plain fd (plusPtr buf n) (size - n)
          case r of
            0 -> return (n, True)
            _ -> loop (n + r)

