{-# OPTIONS_GHC -cpp -optP-P #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Char Memory -> Char Memory transformer which performs newline
   normalization.
-}
module System.IO.SSC.TextMode
  ( iTextMode
  , oTextMode
  , ioTextMode
  , iorTextMode
  , iTextModeWithSize
  , oTextModeWithSize
  , ioTextModeWithSize
  , iorTextModeWithSize
  , ITextMode
  , OTextMode
  , IOTextMode
  , IORTextMode
  ) where

import System.IO.SSC.Classes
import System.IO.SSC.BufferSize

# ifdef mingw32_HOST_OS
import Data.Word
import Foreign.Storable
import System.IO.SSC.Convert
import System.IO.SSC.Buffer

normalizer :: Converter Word8 Word8
normalizer iptr isize optr osize = loop 0 0
  where
    loop read written
      | read == isize || written == osize = return (read, written)
      | otherwise = do
        c <- peekElemOff iptr read
        if c /= 0x0d
          then do
            pokeElemOff optr written c
            loop (read + 1) (written + 1)
          else do
            let read' = read + 1
            if read' == isize
              then return (read, written)
              else do
                c2 <- peekElemOff iptr read'
                if c2 == 0x0a
                  then do
                    pokeElemOff optr written 0x0a
                    loop (read' + 1) (written + 1)
                  else let written' = written + 1 in
                    if written' == osize
                      then return (read, written)
                      else do
                        pokeElemOff optr written 0x0d
                        pokeElemOff optr written' c2
                        loop (read' + 1) (written' + 1)

denormalizer :: Converter Word8 Word8
denormalizer iptr isize optr osize = loop 0 0
  where
    loop read written
      | read == isize || written == osize = return (read, written)
      | otherwise = do
        c <- peekElemOff iptr read
        if c /= 0x0a
          then do
            pokeElemOff optr written c
            loop (read + 1) (written + 1)
          else do
            let written' = written + 1
            if written' == osize
              then return (read, written)
              else do
                pokeElemOff optr written 0x0d
                pokeElemOff optr written' 0x0a
                loop (read + 1) (written' + 1)

type ITextMode d pos a m = IBuffer d (ConversionPos pos) Word8 (IConversionBlock d pos m a Word8 Word8)
type OTextMode d pos a m = OBuffer d (ConversionPos pos) Word8 (OConversionBlock d pos m a Word8 Word8)
type IOTextMode d pos a m = IOBuffer d (ConversionPos pos) Word8 (IOConversionBlock d pos m a Word8 Word8)
type IORTextMode d pos a m = IORBuffer d (ConversionPos pos) Word8 (IOConversionBlock d pos m a Word8 Word8)

iTextModeWithSize :: Int -> m -> IO (ITextMode d pos a m)
iTextModeWithSize size mem = iConversionBlock normalizer mem >>= iSizedBuffer size

oTextModeWithSize :: Int -> m -> IO (OTextMode d pos a m)
oTextModeWithSize size mem = oConversionBlock denormalizer mem >>= oSizedBuffer size

ioTextModeWithSize :: Int -> m -> IO (IOTextMode d pos a m)
ioTextModeWithSize size mem = ioConversionBlock normalizer denormalizer mem >>= ioSizedBuffer size

iorTextModeWithSize :: (Seekable pos m) => Int -> m -> IO (IORTextMode d pos a m)
iorTextModeWithSize size mem = ioConversionBlock normalizer denormalizer mem >>= iorSizedBuffer size

# else /* defined mingw32_HOST_OS */

type ITextMode d pos a m = m
type OTextMode d pos a m = m
type IOTextMode d pos a m = m
type IORTextMode d pos a m = m

iTextModeWithSize :: Int -> m -> IO (ITextMode d pos a m)
iTextModeWithSize size mem = return mem 

oTextModeWithSize :: Int -> m -> IO (OTextMode d pos a m)
oTextModeWithSize size mem = return mem 

ioTextModeWithSize :: Int -> m -> IO (IOTextMode d pos a m)
ioTextModeWithSize size mem = return mem 

iorTextModeWithSize :: (Seekable pos m) => Int -> m -> IO (IORTextMode d pos a m)
iorTextModeWithSize size mem = return mem 

# endif

iTextMode :: m -> IO (ITextMode d pos a m)
iTextMode m = iTextModeWithSize defaultBufferSize m

oTextMode :: m -> IO (OTextMode d pos a m)
oTextMode m = oTextModeWithSize defaultBufferSize m

ioTextMode :: m -> IO (IOTextMode d pos a m)
ioTextMode m = ioTextModeWithSize defaultBufferSize m

iorTextMode :: (Seekable pos m) => m -> IO (IORTextMode d pos a m)
iorTextMode m = iorTextModeWithSize defaultBufferSize m
