# include <iconv.h>
# include <stddef.h>
# include <stdlib.h>
# include <errno.h>

struct iconv_d
{
  iconv_t t;
  int last_read;
  int last_written;
};

void *iconv_block_new_iconv_d(const char *from, const char *to)
{
  struct iconv_d *ret;
  iconv_t t = iconv_open(to, from);
  if(t == (iconv_t)-1)
    return NULL;
  ret = malloc(sizeof(struct iconv_d));
  if(!ret)
    return NULL;
  ret->t = t;
  return ret;
}

void iconv_block_delete_iconv_d(void *p)
{
  struct iconv_d *d = (struct iconv_d *)p;
  iconv_close(d->t);
  free(d);
}

int iconv_block_convert(void *p, char *from, size_t from_size, char *to, size_t to_size)
{
  struct iconv_d *d = (struct iconv_d *)p;
  char *inp = from, *outp = to;
  int r = iconv(d->t, &inp, &from_size, &outp, &to_size);
  d->last_read = inp - from;
  d->last_written = outp - to;
  if(r == -1 && errno != E2BIG && errno != EINVAL)
    return -1;
  return 0;
}

int iconv_block_last_read(void *p)
{
  struct iconv_d *d = (struct iconv_d *)p;
  return d->last_read;
}

int iconv_block_last_written(void *p)
{
  struct iconv_d *d = (struct iconv_d *)p;
  return d->last_written;
}
